/*
 *      This program is based on GREP from the following sources:
 *
 * The  information  in  this  document  is  subject  to  change
 * without  notice  and  should not be construed as a commitment
 * by Digital Equipment Corporation or by DECUS.
 *
 * Neither Digital Equipment Corporation, DECUS, nor the authors
 * assume any responsibility for the use or reliability of  this
 * document or the described software.
 *
 *      Copyright (C) 1980, DECUS
 *
 *
 * General permission to copy or modify, but not for profit,  is
 * hereby  granted,  provided that the above copyright notice is
 * included and reference made to  the  fact  that  reproduction
 * privileges were granted by DECUS.
 *
 */

#include "stdio.h"
/*
 * change.
 *
 */

#define QUOTE '@'
#define BLANK '#'
#define ESCAPE '\\'

char    *documentation[] = {
"change searches a file for a given pattern and replaces it with ",
" another pattern.  Execute by",
"   chg [flags] regular_expression substitue_expression file_list",
"",
"Flags are single characters preceeded by '-':",
"    -e  delete empty lines",
"    -o  only output lines with matches",
"",
"The file_list is a list of files (wildcards are acceptable).",
"If mulitple files are given, all are concatenated on the std output.",
"If no files are given the std input is read.",
"",
0 };


char    *patdoc[] = {
"The regular_expression defines the pattern to search for.  Upper- and",
"lower-case are always ignored.  Blank lines never match.  The expression",
"should be quoted to prevent file-name translation.",
"x      An ordinary character (not mentioned below) matches that character.",
"'\\'   The backslash escapes tab, backspace, formfeed, newline, and",
"'          carriage return with \\t, \\b, \\f, \\n, and \\r.",
"'@'    The at-sign quotes any character.  \"@$\" matches a dollar-sign.",
"'^'    A circumflex at the beginning of an expression matches the",
"       beginning of a line.",
"'$'    A dollar-sign at the end of an expression matches the end of a line.",
"'.'    A period matches any character except \"new-line\".",
"'#'    An \"pound-sign\" matches one blank.",
"':a'   A colon matches a class of characters described by the following",
"':d'     character.  \":a\" matches any alphabetic, \":d\" matches digits,",
"':n'     \":n\" matches alphanumerics, \": \" matches spaces, tabs, and",
"':#'     other control characters, such as new-line.",
"'*'    An expression followed by an asterisk matches zero or more",
"       occurrances of that expression: \"fo*\" matches \"f\", \"fo\"",
"       \"foo\", etc.",
"'+'    An expression followed by a plus sign matches one or more",
"       occurrances of that expression: \"fo+\" matches \"fo\", etc.",
"'-'    An expression followed by a minus sign optionally matches",
"       the expression.",
"'[]'   A string enclosed in square brackets matches any character in",
"       that string, but no others.  If the first character in the",
"       string is a circumflex, the expression matches any character",
"       except \"new-line\" and the characters in the string.  For",
"       example, \"[xyz]\" matches \"xx\" and \"zyx\", while \"[^xyz]\"",
"       matches \"abc\" but not \"axb\".  A range of characters may be",
"       specified by two characters separated by \"-\".  Note that,",
"       [a-z] matches alphabetics, while [z-a] never matches.",
"'{}'   Group regular expressions into one pattern element: f{oo}+",
"       will only match an 'f' followed by an even number of 'o's.",
"       Braces may be nested.  They are not allowed in square brackets.",
"The concatenation of regular expressions is a regular expression.",
"",
"",
0};

char    *subdoc[] = {
"The substitution_expression which which to replace the matched pattern.",
"Upper- and lower-case are significant in substitution.",
"x      An ordinary character (not mentioned below) is inserted as-is.",
"'\\'   The backslash escapes tab, backspace, formfeed, newline, and",
"'          carriage return with \\t, \\b, \\f, \\n, and \\r.",
"'@'    The at-sign quotes any character.  \"@&\" matches a 'and-sign'.",
"'.'    Represents the null string.  Used to delete the whole pattern.",
"'#'    An \"pound-sign\" matches one blank.",
"'&'    Represents the entire matched pattern.  If the pattern was",
"       change f.o x& , then an x would be added prior to the area ",
"       matched (before the f).  It is the same as &0 (see below).",
"'&1' through '&80'   These represent the nth field in braces.  &0 is ",
"       a special case which represents the whole matched pattern.  For",
"       example:  chg {:d:d}{:a} &2&1 , would move the alpha character",
"       in front of the numbers.  A brace in a *, +, or - is counted as",
"       one field at its maximum match.  It may be empty.  To figure the ",
"       number for a field, start at the left of the pattern with 0 and ",
"       add one for each brace encountered. (&0 is the same as & alone).",
""
0};

#define MSDOS 1
#define DeSmet 1


#ifndef stdin
#define stdin STDIN
#define stdout STDOUT
#define stderr STDERR
#endif

#define LMAX    512
#define PMAX    256
#define OMAX    512

#define CHAR    1
#define BOL     2
#define EOL     3
#define ANY     4
#define CLASS   5
#define NCLASS  6
#define STAR    7
#define PLUS    8
#define MINUS   9
#define ALPHA   10
#define DIGIT   11
#define NALPHA  12
#define PUNCT   13
#define RANGE   14
#define ENDPAT  15
#define BRACE   16
#define ENDBRACE  17

#define ANDMAX 80       /* may only be two digits long */
#define AND    18

struct {               /* store & matches here */
    char *bstart;
    char *bend;
} braces[ANDMAX+1];
int brndx = 0;         /* index to braces arrary */

int     nfile;
int     eflag = 0;
int     oflag = 0;

int     debug   =       0;         /* Set for debug code      */

char    *pp, *subp;     /* pattern pointers */

char    file_name[81];

char    lbuf[LMAX];
char    pbuf[PMAX];
char    sbuf[PMAX];
char    obuf[OMAX];
char	wrkbuf[80];

/*******************************************************/

main(argc, argv)
char *argv[];
{
   register char   *p;
   register int    c, i;
   int             gotpattern, gotsubpat;
   int             gotcha;
   char  *fexpnd();

   FILE            *f;

   if (argc <= 1)
      usage("No arguments");
   if (argc == 2 && argv[1][0] == '?' && argv[1][1] == 0) {
      help(documentation);
      help(patdoc);
      help(subdoc);
      return;
      }
   nfile = argc-1;
   gotsubpat = gotpattern = 0;
   for (i=1; i < argc; ++i) {
      p = argv[i];
      if (*p == '-') {
         ++p;
         while (c = *p++) {
            switch(tolower(c)) {

            case '?':
               help(documentation);
               break;

            case 'D':
            case 'd':
               ++debug;
               break;

            case 'E':
            case 'e':
               ++eflag;
               break;

            case 'O':
            case 'o':
               ++oflag;
               break;

            default:
               usage("Unknown flag");
            }
         }
         argv[i] = 0;
         --nfile;
      } else if (!gotpattern) {
         compile(p);
         argv[i] = 0;
         ++gotpattern;
         --nfile;
      } else if (!gotsubpat) {
         compsub(p);
         argv[i] = 0;
         ++gotsubpat;
         --nfile;
      }
   }
   if (!gotpattern)
      usage("No search pattern");
   if (!gotsubpat)
      usage("No substitution pattern");
   if (nfile == 0)
      change(stdin, 0);
   else {
      for (i=1; i < argc; ++i) {
	     if (argv[i])
		     while (p = fexpnd(argv[i])) 
        		 if (p = argv[i]) 
		            if ((f=fopen(p, "r")) == NULL)
        		       cant(p);
		            else {
        		       change(f, p);
		               fclose(f);
        		    }
	   }
	}
}

/*******************************************************/

usage(s)
char    *s;
{
   puts("?CHANGE-E-");
   puts(s);
   puts("\n");
   puts("Usage: chg [-eo] pattern subst [file ...].  chg ? for help\n");
   exit(1);
}



/*******************************************************/


compile(source)
char       *source;   /* Pattern to compile         */
/*
 * Compile the pattern into global pbuf[]
 */
{
   register char  *s;         /* Source string pointer     */
   register int   c;          /* Current character         */
   int            o;          /* Temp                      */
   char           *pcompil();  /* Compile braced group     */

   pp = pbuf;
   s = source;
   if (debug) {
      puts("Pattern = \"");
      puts(s);
      puts("\"\n");
     }
   s = pcompil(s);
   if (*(--s) == '}')
        badpat("Unmatched close brace",source,s);
}

/*******************************************************/

char *
pcompil(source)
char       *source;   /* Pattern to compile         */
/*
 * Compile (recursively) the pattern into global pbuf[]
 */
{
   register char  *s;         /* Source string pointer     */
   char           *lp;        /* Last pattern pointer      */
   register int   c;          /* Current character         */
   int            o;          /* Temp                      */
   char           *spp;       /* Save beginning of pattern */
   char           *cclass();  /* Compile class routine     */
   char           *sbrace();  /* Compile braced group      */
   char           spchr();     /* Translate char to escaped char */

   s = source;
   while (c = *s++) {
      /*
       * STAR, PLUS and MINUS are special.
       */
      if (c == '*' || c == '+' || c == '-') {
         if (pp == pbuf ||
              (o=pp[-1]) == BOL ||
              o == EOL ||
              o == STAR ||
              o == PLUS ||
              o == MINUS)
            badpat("Illegal occurrance op.", source, s);
         store(ENDPAT);
         store(ENDPAT);
         spp = pp;               /* Save pattern end     */
         while (--pp > lp)       /* Move pattern down    */
            *pp = pp[-1];        /* one byte             */
         *pp =   (c == '*') ? STAR :
            (c == '-') ? MINUS : PLUS;
         pp = spp;               /* Restore pattern end  */
         continue;
      }
      /*
       * All the rest.
       */
      lp = pp;         /* Remember start       */
      switch(c) {

      case BLANK:
         store(CHAR);
        store(' ');
         break;

      case '^':
         store(BOL);
         break;

      case '$':
         store(EOL);
         break;

      case '.':
         store(ANY);
         break;

      case '}':         /* brace pairs group pattern strings */
         return(s);

      case '{':         /* brace pairs group pattern strings */
         s = sbrace(source, s);
         break;

      case '[':
         s = cclass(source, s);
         break;

      case ':':
         if (*s) {
            c = *s++;
            switch(tolower(c)) {

            case 'a':
            case 'A':
               store(ALPHA);
               break;

            case 'd':
            case 'D':
               store(DIGIT);
               break;

            case 'n':
            case 'N':
               store(NALPHA);
               break;

            case BLANK:
               store(PUNCT);
               break;

            default:
               badpat("Unknown : type", source, s);

            }
            break;
         }
         else    badpat("No : type", source, s);

      case QUOTE:
         if (*s) {
                c = *s++;
                store(CHAR);
                store(tolower(c));
        }
        break;

      case ESCAPE:
         if (*s) {
                store(CHAR);
                store(spchr(tolower(c=*s++)));
         }
        break;

      default:
         store(CHAR);
         store(tolower(c));
      }
   }
   store(ENDPAT);
   store(0);                /* Terminate string     */
   if (debug) {
      for (lp = pbuf; lp < pp;) {
         if ((c = (*lp++ & 0377)) < ' ') {
            puts(itoa(c, wrkbuf));
         }
         else {
			wrkbuf[0] = c;
			wrkbuf[1] = '\0';
			puts(wrkbuf);
		}
     }
     puts("\n");
   }
   return(s);
}

/*******************************************************/

compsub(source)
char       *source;   /* Pattern to compile         */
/*
 * Compile the substitution patter into global sbuf[]
 */
{
   register char  *s;         /* Source string pointer     */
   register int   c, c2;      /* Current character         */
   char            *lp;       /* temp string pointer       */
   int            o;          /* Temp                      */
   char           spchr();     /* Translate char to escaped char */

   s = source;
   subp = sbuf;

   while (c = *s++) {

      switch(c) {

      case BLANK:
         stores(CHAR);
         stores(' ');
         break;

      case '.':
         break;         /* ignore it */

      case '&':
         if ((o = atoi(s)) > brndx)     /* brndx is set by pcompil */
              badpat("&nn > max", source, s);
         stores(AND);
         stores(o);
         while (isdigit(*s))
             s++;               /* skip over number */
         break;

      case QUOTE:
         if (*s) {
                c = *s++;
                stores(CHAR);
                stores(tolower(c));
        }
        break;

      case ESCAPE:
         if (*s) {
                stores(CHAR);
                stores(spchr(tolower(c=*s++)));
        }
        break;

      default:
         stores(CHAR);
         stores(tolower(c));
      }
   }
   stores(ENDPAT);
   stores(0);                /* Terminate string     */
   if (debug) {
      puts("sub = ");
      puts(itoa(subp - sbuf, wrkbuf));
      puts("\n");
      for (lp = sbuf; lp < subp;) {
         if ((c = (*lp++ & 0377)) < ' ') 
            puts(itoa(c, wrkbuf));
         else {
			wrkbuf[0] = c;
			wrkbuf[1] = '\0';
			puts(wrkbuf);
		}
     }
     puts("\n");
   }
   return(s);
}

/*******************************************************/

char spchr(inchar)
char inchar;
{
        switch(inchar) {

                case 't':
                        return('\t');
                case 'b':
                        return('\b');
                case 'f':
                        return('\f');
                case 'n':
                        return('\n');
                case 'r':
                        return('\r');
                default:
                        return(inchar);
                }
}

/*******************************************************/

char *
cclass(source, src)
char       *source;   /* Pattern start -- for error msg.      */
char       *src;      /* Class start           */
/*
 * Compile a class (within [])
 */
{
   register char   *s;        /* Source pointer    */
   register char   *cp;       /* Pattern start     */
   register int    c;         /* Current character */
   int             o;         /* Temp              */
   char           spchr();     /* Translate char to escaped char */

   s = src;
   o = CLASS;
   if (*s == '^') {
      ++s;
      o = NCLASS;
   }
   store(o);
   cp = pp;
   store(0);                          /* Byte count      */
   while ((c = *s++) && c!=']') {
      if (c == QUOTE) {                /* Store quoted char    */
         if ((c = *s++) == '\0')      /* Gotta get something  */
            badpat("Class terminates badly", source, s);
         else    store(tolower(c));
      }
      else if (c == ESCAPE) {
                if ((c = *s++) == '\0')
                        badpat("Invalid Escape sequence", source, s);
                else
                        store(spchr(c));
        }
      else if (c == '-' &&
            (pp - cp) > 1 && *s != ']' && *s != '\0') {
         c = pp[-1];             /* Range start     */
         pp[-1] = RANGE;         /* Range signal    */
         store(c);               /* Re-store start  */
         c = *s++;               /* Get end char and*/
         store(tolower(c));      /* Store it        */
      }
      else {
         store(tolower(c));      /* Store normal char */
      }
   }
   if (c != ']')
      badpat("Unterminated class", source, s);
   if ((c = (pp - cp)) >= 256)
      badpat("Class too large", source, s);
   if (c == 0)
      badpat("Empty class", source, s);
   *cp = c;
   return(s);
}


/*******************************************************/

char *
sbrace(source, src)
char       *source;   /* Pattern start -- for error msg.      */
char       *src;      /* Bracketed area start  */
/*
 * Compile a pattern withing {} (a grouping)
 */
{
   register char   *s;        /* Source pointer    */
   register int    c;         /* Current character */
   int             o;         /* Temp              */

   s = src;
   o = BRACE;
   store(o);

   if (++brndx > ANDMAX)      /* save max used for compsub */
       badpat("Too many braces",source,s);
   s = pcompil(s);         /* call compile to finish pattern */
   if (s[-1] == '}')    /* set up end of group */
        store(ENDBRACE);
   else
        badpat("Unmatched brace",source,s);
   return(s);
}

/*******************************************************/

store(op)
{
   if (pp >= &pbuf[PMAX])
      error("Pattern too complex\n","");
   *pp++ = op;
}

/*******************************************************/

stores(op)
{
   if (subp >= &sbuf[PMAX])
      error("Pattern too complex\n","");
   *subp++ = op;
}


/*******************************************************/

badpat(message, source, stop)
char  *message;       /* Error message */
char  *source;        /* Pattern start */
char  *stop;          /* Pattern end   */
{
   register int    c, lp;

   if (debug) {
      for (lp = pbuf; lp < pp;) {
         if ((c = (*lp++ & 0377)) < ' ') 
            puts(itoa(c, wrkbuf));
         else {
			wrkbuf[0] = c;
			wrkbuf[1] = '\0';
			puts(wrkbuf);
		}
     }
     puts("\n");
   }
   puts("-CHG-E-");
   puts(message);
   puts(" pattern is\"");
   puts(source);
   puts("\"\n");
   puts("-CHG-E-Stopped at byte ");
   puts(itoa(stop-source, wrkbuf));
   puts(" '");
   wrkbuf[0] = c;
   wrkbuf[1] = '\0';
   puts(wrkbuf);
   puts("' \n");
   error("?CHG-E-Bad pattern\n","");
}



/*******************************************************/

change(fp, fn)
FILE       *fp;       /* File to process            */
char       *fn;       /* File name (for -f option)  */
/*
 * Scan the file for the pattern in pbuf[]
 */
{
   register int lno, indx;
   int c, m;
   char *l, *match(), *subst();

   lno = 0;
   while (fgets(lbuf, LMAX, fp)) {
      ++lno;
      m = 0;
      l = lbuf;
      while (l != 0)
          if (l = match(l)) {
              m++;
              l = subst();    /* copies to obuf with substitution */
              strncpy(lbuf,obuf,OMAX); /* and replace for multiple subs */
          }
      if (m || !oflag)                     /* write line? */
         if ((lbuf[0] != '\n') || !eflag)  /* skip empty lines? */
             puts(lbuf);
   }
}


/*******************************************************/

char *
match(lstart)
char *lstart;
/*
 * Match the current line (in lbuf[]), return l if it does.
 */
{
   register char   *l;        /* Line pointer       */
   char *e;                   /* end of pattern pointer */
   int i;
   char *pmatch();

   for (l = lstart; *l; l++) {
      brndx = 0;
      if (e = pmatch(l, pbuf)) {
         braces[0].bstart = l;  /* record match range for subst */
         braces[0].bend = e;
         return(e);
      }
   }
   return(0);
}

/*******************************************************/

char *
pmatch(line, pattern)
char               *line;     /* (partial) line to match      */
char               *pattern;  /* (partial) pattern to match   */
{
   char   *l;        /* Current line pointer         */
   char   *p;        /* Current pattern pointer      */
   char   c;         /* Current character            */
   char   *e;        /* End for STAR and PLUS match  */
   int    op;        /* Pattern operation            */
   int    n;         /* Class counter                */
   int    savbrace;  /* Brace reset point            */
   char   *are;      /* Start of STAR match          */

   l = e = line;
   if (debug > 1) {
      puts("pmatch(\"");
      puts(line);
      puts("\")\n");
   }
   p = pattern;
   while ((op = *p++) != ENDPAT && op != ENDBRACE) {
      if (debug > 1) {
         puts("byte[");
         puts(itoa(l-lbuf, wrkbuf));
         puts("] = 0");
         puts(itoa(*l, wrkbuf));
		 puts(" '");
		 wrkbuf[0] = *l;
		 wrkbuf[1] = '\0';
		 puts("', op = 0");
		 puts(itoa(op, wrkbuf));
		 puts("\n");
	  }
      switch(op) {

      case CHAR:
         if (tolower(*l++) != *p++)
            return(0);
         break;

      case BOL:
         if (l != lbuf)
            return(0);
         break;

      case EOL:
         if (*l != '\0')
            return(0);
         break;

      case ANY:
         if (*l++ == '\0')
            return(0);
         break;

      case DIGIT:
         if ((c = *l++) < '0' || (c > '9'))
            return(0);
         break;

      case ALPHA:
         c = tolower(*l++);
         if (c < 'a' || c > 'z')
            return(0);
         break;

      case NALPHA:
         c = tolower(*l++);
         if (c >= 'a' && c <= 'z')
            break;
         else if (c < '0' || c > '9')
            return(0);
         break;

      case PUNCT:
         c = *l++;
         if (c == 0 || c > ' ')
            return(0);
         break;

      case BRACE:
         savbrace = brndx;      /* save for backout */
         if (++brndx <= ANDMAX)
             braces[brndx].bstart = l;  /* record start of match */
          else
             error("braces overflowed\n","");

         e = pmatch(l, p);      /* match braced group */

         for(n=1; n; p++)       /* skip matched group */
             if (*p == ENDBRACE)
                 n--;           /* end of some group */
             else if (*p == BRACE)
                 n++;           /* start of another group */

         if (e) {
             braces[savbrace+1].bend = e;   /* update end of match */
             l = e;      /* if match, update area */
         }
         else {
            brndx = savbrace;   /* ignore inner matched braces */
            return(0);  /* else, return no match */
         }
         break;

      case CLASS:
      case NCLASS:
         c = tolower(*l++);
         n = *p++ & 0377;
         do {
            if (*p == RANGE) {
               p += 3;
               n -= 2;
               if (c >= p[-2] && c <= p[-1])
                  break;
            }
            else if (c == *p++)
               break;
         } while (--n > 1);
         if ((op == CLASS) == (n <= 1))
            return(0);
         if (op == CLASS)
            p += n - 2;
         break;

      case MINUS:
         e = pmatch(l, p);       /* Look for a match    */
         while (*p++ != ENDPAT); /* Skip over pattern   */
         if (e)                  /* Got a match?        */
            l = e;               /* Yes, update string  */
         break;                  /* Always succeeds     */

      case PLUS:                 /* One or more ...     */
         if ((l = pmatch(l, p)) == 0)
            return(0);           /* Gotta have a match  */
      case STAR:                 /* Zero or more ...    */
         are = l;                /* Remember line start */
         while (*l && (e = pmatch(l, p)))
            l = e;               /* Get longest match   */
         while (*p++ != ENDPAT); /* Skip over pattern   */
         while (l >= are) {      /* Try to match rest   */
            if (e = pmatch(l, p))
               return(e);
            --l;                 /* Nope, try earlier   */
         }
         return(0);              /* Nothing else worked */

      default:
         puts("Bad op code ");
         puts(itoa(op, wrkbuf));
         puts("\n");
         error("Cannot happen -- match\n","");
      }
   }
   return(l);
}

/*******************************************************/

char *
subst()
{
   char   *l;        /* Current input line pointer   */
   char   *o;        /* Current output line pointer  */
   char   *t;        /* Temporary pattern pointer    */
   char   *p;        /* Current pattern pointer      */
   char   *le;       /* Save area for return         */
   char   c;         /* Current character            */
   int    op;        /* Pattern operation            */
   int    ndx;       /* Matched area index           */

   l = lbuf;
   o = obuf;
   p = sbuf;

   while (l < braces[0].bstart)
       *o++ = *l++;

   while ((op = *p++) != ENDPAT) {

      switch(op) {

      case CHAR:
         *o++ = *p++;   /* use next char */
         break;

      case AND:
         ndx = (int)*p++;
         if (ndx <= brndx)  /* if not matched, use null string */
             for(t = braces[ndx].bstart; t < braces[ndx].bend; )
                *o++ = *t++;
         break;


      default:
         puts("Bad op code ");
         puts(itoa(op, wrkbuf));
         puts("\n");
         error("Cannot happen -- subst\n","");
      }
   }

   l = braces[0].bend;  /*  skip matched area on input stream */
   le = o - obuf + lbuf;  /* save scan start in input buffer for return */
   while (*o++ = *l++)  /*  and copy rest of input to output */
	   	;

   return(le);
}

